use futures_core::stream::Stream;
use futures_core::task::{Waker, Poll};
use pin_utils::{unsafe_pinned, unsafe_unpinned};
use std::any::Any;
use std::pin::Pin;
use std::panic::{catch_unwind, UnwindSafe, AssertUnwindSafe};
use std::prelude::v1::*;

/// Stream for the `catch_unwind` combinator.
///
/// This is created by the `Stream::catch_unwind` method.
#[derive(Debug)]
#[must_use = "streams do nothing unless polled"]
pub struct CatchUnwind<St: Stream> {
    stream: St,
    caught_unwind: bool,
}

impl<St: Stream + UnwindSafe> CatchUnwind<St> {
    unsafe_pinned!(stream: St);
    unsafe_unpinned!(caught_unwind: bool);

    pub(super) fn new(stream: St) -> CatchUnwind<St> {
        CatchUnwind { stream, caught_unwind: false }
    }
}

impl<St: Stream + UnwindSafe> Stream for CatchUnwind<St>
{
    type Item = Result<St::Item, Box<dyn Any + Send>>;

    fn poll_next(
        mut self: Pin<&mut Self>,
        waker: &Waker,
    ) -> Poll<Option<Self::Item>> {
        if *self.as_mut().caught_unwind() {
            Poll::Ready(None)
        } else {
            let res = catch_unwind(AssertUnwindSafe(|| {
                self.as_mut().stream().poll_next(waker)
            }));

            match res {
                Ok(poll) => poll.map(|opt| opt.map(Ok)),
                Err(e) => {
                    *self.as_mut().caught_unwind() = true;
                    Poll::Ready(Some(Err(e)))
                },
            }
        }
    }
}
